{.PW132}
{.IN+}
{.HE RULES.PAS                                          Page #}
{$V-,R+}
PROGRAM rule_builder ;

(* Copyright (c) 1987 - Knowledge Garden Inc.
                        473A Malden Bridge Rd.
                        RD #2
                        Nassau, NY 12123       *)


(* RULES builds rule based systems by carrying on a dialog with the user.
   It is described in the November 1987 AI apprentice column in AI Expert
   magazine. THe list processing routine are from VT-LISP which was described
   in the April and May 1987 issues of AI Expert.

   This program has been tested using Turbo ver 3.02 on an IBM PC. It has
   been run under DOS 3.2.

   RULES produces a knowledge base which can be used with MicroExpert.
   MicroExpert is an expert system shell written in Turbo Pascal for the
   IBM PC and Apple II. It is available for $49.95 and comes with complete
   source code. It can be order by writing to :
        McGraw-Hill Book Company
        P.O. Box 400
        Hightstown, NJ 08520

   Or calling 1-800-628-004 or in New York state 212/512-2999.

   It shouldn't be hard to modify the program to produce rules for other
   similar systems. Just chnage the routine, write_rules.

   We would be pleased to hear your comments, good or bad, or any applications
   and modifications of the program. Contact us at:

     AI Expert
     CL Publications Inc.
     500 Howard St.
     San Francisco, CA 94105

   or on the AI Expert BBS on Compuserv.
   Our id is BillandBev Thompson ,[76703,4324].
   You can also contact us on BIX, our id is bbt.

   Bill and Bev Thompson    *)

 CONST
  bell = ^G ;
  back_space = ^H ;
  tab = ^I ;

 TYPE
  counter = 0 .. maxint ;
  string132 = string[132] ;
  string80 = string[80] ;
  string255 = string[255] ;
  text_file = text ;
  char_set = SET OF char ;
  node_type = (cons_node,symbol,number,free_node) ;
  s_expr = ^node ;
  node = RECORD
          in_use : boolean ;
          CASE tag : node_type OF
           cons_node : (car_ptr : s_expr ;
                        cdr_ptr : s_expr) ;
           symbol    : (string_data : string80) ;
           number    : (num_data : real) ;
           free_node : (next_free : s_expr ;
                        block_cnt : counter) ;
          END ;

(* node is the basic allocation unit for lists. The fields are used as
   follows:

    in_use     - in_use = false tells the garbage collector that this node
                 is available for re-use.
    tag        - which kind of node this is.
    cons_node  - cons_nodes consist of two pointers. one to the head (first item)
                 the other to the rest of the list. They are the "glue" which
                 holds the list together. The list (A B C) would be stored as
                   -------         --------          --------
                   | .| . |----->  |  .| . |------> |  .| . |---> NIL
                   --|-----         --|------        --|-----
                     |                |                |
                     V                V                V
                     A                B                C

                 The boxes are the cons nodes, the first part of the box
                 holds the car pointer, then second contains the cdr pointer.
    symbol     - holds string values, we don't actually use the entire 80
                 characters in most cases.
    number     - used for storage of numbers. All numbers are implemented as
                 reals. This is inefficient, but relatively easy in Turbo
                 Pascal.
    free_node  - the garbage collector gathers all unused nodes and puts
                 them on a free list. It also compacts the free space into
                 contiguous blocks. next_free points to the next free block.
                 block_cnt contains a count of the number of contiguous 8 byte free
                 blocks which follow this one.



    Note: we often allocate a new node for each atom, instead of a pointer to an
          existing string in the heap. This slows down comparisons,
          because you have to compare strings instead of pointers, but
          speeds up allocation. We've tried it both ways and there seems
          to be no effect on small programs, but if you decide to
          expand this program you should take a long hard look at all
          of the allocation routines and improve them. *)


 VAR
  main_goal,free,rule_list,ques_list,initial_heap,saved_list : s_expr ;
  total_free : real ;
  rule_file : text_file ;
  rule_count : counter ;

(* Variables - These are the important global variables:
     total_free -  a count of the total amount of free memory on the
                   free list.
     rule_list  - the set of rules produced by build_rule. Its structure
                  is described in the build_rule routine.
     free        - a linked list of free nodes. Memory is allocated from
                   from here if possible before getting memory from
                   the heap. This list is built by the garbage collector.
     inital_heap - a pointer to the bottom of the heap
     saved_list  - a list of all nodes which must absolutely not be
                   reclaimed by the garbage collector.
     ques_list   - the list of attributes, their corresponding values and
                   the text of the questions. Its structure is described in
                   the write_question routine
     main_goal   - the goal for the rule set.
     rule_file   - this file will contain the rules
     rule_count  - just a count of rules as they are written, useful for
                   MicroExpert because it needs rule numbers. *)


(* ----------------------------------------------------------------------
        Utility Routines
   ---------------------------------------------------------------------- *)

(* Some of these routines may not actually be used, we lifted them from
   VT-LISP *)

 FUNCTION open(VAR f : text_file ; f_name : string80) : boolean ;
  (* open a file - returns true if the file exists and was opened properly
     f      - file pointer
     f_name - external name of the file *)
  BEGIN
   assign(f,f_name) ;
   (*$I- *)
   reset(f) ;
   (*$I+ *)
   open := (ioresult = 0) ;
  END ; (* open *)


 FUNCTION is_console(VAR f : text_file) : boolean ;
  (* return true if f is open on the system console
     for details of fibs and fib_ptrs see the Turbo Pascal ver 3.0 reference
     manual chapter 20. This should work under CP/M-86 or 80, but we haven't
     tried it. *)
  TYPE
   fib = ARRAY [0 .. 75] OF byte ;
  VAR
   fib_ptr : ^fib ;
   dev_type : byte ;
  BEGIN
   fib_ptr := addr(f) ;
   dev_type := fib_ptr^[2] AND $07 ;
   is_console := (dev_type = 1) OR (dev_type = 2) ;
  END ; (* is_console *)


 PROCEDURE strip_leading_blanks(VAR s : string80) ;
  BEGIN
   IF length(s) > 0
    THEN
     IF (s[1] = ' ') OR (s[1] = tab)
      THEN
       BEGIN
        delete(s,1,1) ;
        strip_leading_blanks(s) ;
       END ;
  END ; (* strip_leading_blanks *)


 PROCEDURE strip_trailing_blanks(VAR s : string80) ;
  BEGIN
   IF length(s) > 0
    THEN
     IF (s[length(s)] = ' ') OR (s[length(s)] = tab)
      THEN
       BEGIN
        delete(s,length(s),1) ;
        strip_trailing_blanks(s) ;
       END ;
  END ; (* strip_trailing_blanks *)



 FUNCTION toupper(s : string80) : string80 ;
  (* returns s converted to upper case *)
  VAR
   i : byte ;
  BEGIN
   IF length(s) > 0
    THEN
     FOR i := 1 TO length(s) DO
      s[i] := upcase(s[i]) ;
   toupper := s ;
  END ; (* toupper *)


 FUNCTION toreal(s : string255) : real ;
  (* Converts "s" to a real - ignores non-numeric characters. *)
  VAR
   num : real ;
   code : integer ;
  BEGIN
   strip_trailing_blanks(s) ;
   strip_leading_blanks(s) ;
   IF s = ''
    THEN code := -1
   ELSE IF length(s) = 1
    THEN
     IF s[1] IN ['0' .. '9']
      THEN val(s,num,code)
      ELSE code := -1
   ELSE val(s,num,code) ;
   IF code = 0
    THEN toreal := num
    ELSE toreal := 0 ;
  END ; (* toreal *)


 FUNCTION tointeger(s : string80) : integer ;
  VAR
   num : real ;
   code : integer ;
  BEGIN
   strip_trailing_blanks(s) ;
   strip_leading_blanks(s) ;
   val(s,num,code) ;
   IF code = 0
    THEN
     IF (num < -32768.0) OR (num > 32767.0)
      THEN tointeger := 0
      ELSE tointeger := trunc(num)
    ELSE tointeger := 0 ;
  END ; (* tointeger *)


 FUNCTION is_number(s : string255) : boolean ;
  VAR
   num : real ;
   code : integer ;
  BEGIN
   strip_trailing_blanks(s) ;
   strip_leading_blanks(s) ;
   IF s = ''
    THEN code := -1
   ELSE IF length(s) = 1
    THEN
     IF S[1] IN ['0' ..'9']
      THEN code := 0
      ELSE code := -1
   ELSE val(s,num,code) ;
   is_number := (code = 0) ;
  END ; (* is_number *)


 FUNCTION cardinal(i : integer) : real ;
  VAR
   r : real ;
  BEGIN
   r := i ;
   IF r < 0
    THEN r := r + 65536.0 ;
   cardinal := r ;
  END ; (* cardinal *)


 FUNCTION tag_value(list : s_expr) : node_type ;
  (* returns the value of the tag for a node.     *)
  BEGIN
   IF list = NIL
    THEN tag_value := free_node
    ELSE tag_value := list^.tag ;
  END ; (* tag_value *)


 FUNCTION car(list : s_expr) : s_expr ;
  (* returns a pointer to the first item in the list.
     If the list is empty, it returns NIL.  *)
  BEGIN
   IF list = NIL
    THEN car := NIL
   ELSE IF tag_value(list) = cons_node
    THEN car := list^.car_ptr
   ELSE car := NIL ;
  END ; (* car *)


 FUNCTION cdr(list : s_expr) : s_expr ;
  (* returns a pointer to a list starting at the second item in the list.
     Note - cdr( (a b c) ) points to the list (b c), but
            cdr( ((a b) c d) ) points to the list (c d) .  *)
  BEGIN
   IF list = NIL
    THEN cdr := NIL
    ELSE
     CASE tag_value(list) OF
      cons_node : cdr := list^.cdr_ptr ;
      free_node : cdr := list^.next_free ;
      ELSE        cdr := NIL ;
     END ;
  END ; (* cdr *)


 FUNCTION last(list : s_expr) : s_expr ;
  (* return the last element of a list.
     The last element of an element is nil and the last element of a nil
     list is nil. *)
  BEGIN
   IF cdr(list) = NIL
    THEN last := car(list)
    ELSE last := last(cdr(list)) ;
  END ; (* last *)


 FUNCTION atom(p : s_expr) : boolean ;
  (* Return true if p is a symbolic or numeric atom, otherwise
     it returns false *)
  BEGIN
   IF p = NIL
    THEN atom := false
   ELSE IF tag_value(p) IN [number,symbol]
    THEN atom := true
   ELSE atom := false ;
  END ; (* atom *)


 FUNCTION allocation_size(x : integer) : integer ;
  (* Turbo 3.0 allocates memory in 8 byte blocks, this routine calculates the
     actual number of bytes returned for a request of x bytes.  *)
  BEGIN
   allocation_size := (((x - 1) SHR 3) + 1) SHL 3 ;
  END ; (* allocation_size *)


 FUNCTION node_size : counter ;
  (* calculates the size of a cons node. *)
  BEGIN
   node_size := 2 * sizeof(s_expr) + sizeof(boolean) + sizeof(node_type) ;
  END ; (* node_size *)


 FUNCTION normalize(pt : s_expr) : s_expr ;
  (* returns a normalized pointer. Pointers are 32 bit addresses. The first
     16 bits contain the segment number and the second 16 bits contain the
     offset within the segment. Normalized pointers have offsets in the range
     $0 to $F (0 .. 15)    *)
  VAR
   pt_seg,pt_ofs : integer ;
  BEGIN
   pt_seg := seg(pt^) + (ofs(pt^) SHR 4) ;
   pt_ofs := ofs(pt^) AND $000F ;
   normalize := ptr(pt_seg,pt_ofs) ;
  END ; (* normalize *)


 FUNCTION string_val(list : s_expr) : string80 ;
  (* returns the string pointed to by list. If list points to a number
     node, it returns a string representing that number *)
  TYPE
   real_rec = RECORD
               CASE boolean OF
                true  : (p1 : real) ;
                false : (p2 : ARRAY [0 ..5] OF byte) ;
               END ;
  VAR
   s : string80 ;
   p : real_rec ;

  PROCEDURE strip_trailing_zeros(VAR ss : string80) ;
   BEGIN
    IF ss <> ''
     THEN
      IF ss[length(ss)] = '0'
       THEN
        BEGIN
         delete(ss,length(ss),1) ;
         strip_trailing_zeros(ss) ;
        END ;
   END ; (* strip_trailing_zeros *)

  BEGIN
   IF list = NIL
    THEN string_val := ''
   ELSE IF list^.tag = symbol
    THEN string_val := list^.string_data
   ELSE IF list^.tag = number
    THEN
     WITH list^ DO
      BEGIN
       p.p1 := abs(frac(num_data)) ;
       IF p.p2[0] = 0
        THEN str(num_data : 20 : 0,s)
       ELSE IF p.p2[0] < 112
        THEN str(num_data,s)
       ELSE
        BEGIN
         str(num_data : 20 : 10,s) ;
         strip_trailing_zeros(s) ;
        END ;
       strip_leading_blanks(s) ;
       string_val := s ;
      END
   ELSE string_val := '' ;
  END ; (* string_val *)


 FUNCTION num_val(list : s_expr) : real ;
  (* returns the number pointed to by list. If list points to a string,
     it returns the numerical value of the string.   *)
  BEGIN
   IF list = NIL
    THEN num_val := 0.0
   ELSE IF list^.tag = number
    THEN num_val := list^.num_data
   ELSE IF list^.tag = symbol
    THEN num_val := toreal(list^.string_data)
   ELSE num_val := 0.0 ;
  END ; (* num_val *)


 PROCEDURE get_memory(VAR pt : s_expr ; size : integer) ;
  (* On exit p contains a pointer to a block of allocation_size(size) bytes.
     If possible this routine tries to get memory from the free list before
     requesting it from the heap *)
  VAR
   blks : counter ;
   allocated : boolean ;

  PROCEDURE get_from_free ;
   VAR
    p,q : s_expr ;
   BEGIN
    p := free ;
    q := p ;
    WHILE (p <> NIL) AND (NOT allocated) DO
     IF p^.block_cnt >= (blks - 1)
      THEN
       BEGIN
        pt := normalize(ptr(seg(p^),ofs(p^) + (p^.block_cnt - blks + 1) SHL 3)) ;
        IF p^.block_cnt = (blks - 1)
         THEN
          IF p = free
           THEN free := free^.next_free
           ELSE q^.next_free := p^.next_free
         ELSE p^.block_cnt := p^.block_cnt - blks ;
        allocated := true ;
        total_free := total_free - (blks * 8.0) ;
       END
      ELSE
       BEGIN
        q := p ;
        p := p^.next_free ;
       END ;
   END ; (* get_from_free *)

  BEGIN
   blks := ((size - 1) SHR 3) + 1 ;
   allocated := false ;
   get_from_free ;
   IF NOT allocated
    THEN getmem(pt,blks SHL 3) ;
  END ; (* get_memory *)


 FUNCTION alloc_str(s : string80) : s_expr ;
  (* Allocate storage for a string and return a pointer to the new node.
     This routine only allocates enough storage for the actual number of
     characters in the string plus one for the length. Because of this,
     concatenating anything to the end of a string stored in a symbol node
     will lead to disaster. Copy the string to a new string do the
     concatenation and then allocate a new node.  *)
  VAR
   pt : s_expr ;
  BEGIN
   get_memory(pt,allocation_size(sizeof(node_type) + sizeof(boolean) +
                                 length(s) + 1)) ;
   pt^.tag := symbol ;
   pt^.string_data := s ;
   alloc_str := pt ;
  END ; (* alloc_str *)


 FUNCTION alloc_num(r : real) : s_expr ;
  (* Allocate storage for a number and return a pointer to the new node.
     All numbers are stored as reals. This isn't efficient, but it is
     easy. *)
  VAR
   pt : s_expr ;
  BEGIN
   get_memory(pt,allocation_size(sizeof(node_type) + sizeof(boolean) +
                                 sizeof(real))) ;
   pt^.tag := number ;
   pt^.num_data := r ;
   alloc_num := pt ;
  END ; (* alloc_num *)


 FUNCTION cons(new_node,list : s_expr) : s_expr ;
  (* Construct a list. This routine allocates storage for a new cons node.
     new_node points to the new car of the list. The cdr pointer of the
     new node points to list. This routine adds the new cons node to the
     beginning of the list and returns a pointer to it. The list described
     in the comments at the beginning of the program could be constructed
     as cons(alloc_str('A'),cons(alloc_str('B'),cons(alloc_str('C'),NIL))). *)
  VAR
   p : s_expr ;
  BEGIN
   get_memory(p,allocation_size(node_size)) ;
   p^.tag := cons_node ;
   p^.car_ptr := new_node ;
   p^.cdr_ptr := list ;
   cons := p ;
  END ; (* cons *)


 FUNCTION append_list(list1,list2 : s_expr) : s_expr ;
  (* attach list2 to the end of list1, return a pointer to the
     combined list *)
  BEGIN
   IF list1 = NIL
    THEN append_list := list2
    ELSE append_list := cons(car(list1),append_list(cdr(list1),list2)) ;
  END ; (* append_list *)


 PROCEDURE collect_garbage ;
  (* This routine is specific to Turbo Pascal Ver 3.01
     It depends upon the fact that Turbo allocates memory in 8 byte blocks
     on the PC. If you recompile this program on another system be very
     careful with this routine.
     Garbage collection proceeds in three phases:
      unmark  - free all memory between the initial_heap^ and the current
                top of the heap.
      mark_mem    - mark everything on the saved_list as being in ues.
      release - gather all unmarked blocks and put them on the free list.
     The collector displays a '*' on the screen to let you know it is
      operating.  *)

  FUNCTION lower(p1,p2 : s_expr) : boolean ;
   (* returns true if p1 points to a lower memory address than p2 *)
   BEGIN
    p1 := normalize(p1) ;
    p2 := normalize(p2) ;
    lower := (cardinal(seg(p1^)) < cardinal(seg(p2^))) OR
              ((seg(p1^) = seg(p2^)) AND
              (cardinal(ofs(p1^)) < cardinal(ofs(p2^)))) ;
   END ; (* lower *)

  PROCEDURE mark_mem(list : s_expr) ;
   (* Mark the blocks on list as being in use. Since a node may be on several
      lists at one time, if it is already marked we don't continue processing
      the cdr of the list. *)
   BEGIN
    IF list <> NIL
     THEN
      BEGIN
       IF NOT list^.in_use
        THEN
         BEGIN
          list^.in_use := true ;
          IF list^.tag = cons_node
           THEN
            BEGIN
             mark_mem(car(list)) ;
             mark_mem(cdr(list)) ;
            END ;
         END ;
      END ;
   END ; (* mark_mem *)

  PROCEDURE unmark_mem ;
   (* Go through memory from initial_heap^ to HeapPtr^ and mark each node
      as not in use. The tricky part here is updating the pointer p to point
      to the next cell. *)
   VAR
    p : s_expr ;
    string_base,node_allocation,number_allocation : integer ;
   BEGIN
    string_base := sizeof(node_type) + sizeof(boolean) ;
    p := normalize(initial_heap) ;
    node_allocation := allocation_size(node_size) ;
    number_allocation := allocation_size(sizeof(node_type) + sizeof(boolean) +
                                         sizeof(real)) ;
    WHILE lower(p,HeapPtr) DO
     BEGIN
      p^.in_use := false ;
      CASE p^.tag OF
       cons_node : p := normalize(ptr(seg(p^),ofs(p^) + node_allocation)) ;
       free_node : p := normalize(ptr(seg(p^),ofs(p^) + ((p^.block_cnt + 1) SHL 3))) ;
       number    : p := normalize(ptr(seg(p^),ofs(p^) + number_allocation)) ;
       symbol    : p := normalize(ptr(seg(p^),
                                  ofs(p^) +
                                  allocation_size(string_base +
                                                  length(p^.string_data) + 1))) ;
      END ;
     END ;
   END ; (* unmark_mem *)

  PROCEDURE release_mem ;
   (* This procedure does the actual collection and compaction of nodes.
      This is the slow phase of garbage collection because of all the pointer
      manipulation.  *)
   VAR
    heap_top : s_expr ;
    string_base,node_allocation,string_allocation,block_allocation,
                number_allocation : integer ;

   PROCEDURE free_memory(pt : s_expr ; size : counter) ;
    (* return size bytes pointed to by pt to the free list. If pt points to
       a block next to the head of the free list combine it with the top
       free node. total_free keeps track of the total number of free bytes. *)
    VAR
     blks : counter ;
    BEGIN
     IF pt <> NIL
      THEN
       BEGIN
        blks := ((size - 1) SHR 3) + 1 ;
        pt^.tag := free_node ;
        IF normalize(ptr(seg(pt^),ofs(pt^) + (blks SHL 3))) = free
         THEN
          BEGIN
           pt^.next_free := free^.next_free ;
           pt^.block_cnt := free^.block_cnt + blks ;
           free := pt ;
          END
        ELSE IF normalize(ptr(seg(free^),ofs(free^) +
                      ((free^.block_cnt + 1) SHL 3))) = normalize(pt)
         THEN free^.block_cnt := free^.block_cnt + blks
        ELSE
         BEGIN
          pt^.next_free := free ;
          pt^.block_cnt := blks - 1 ;
          free := pt ;
         END ;
        total_free := total_free + (blks * 8.0) ;
       END ;
    END ; (* free_memory *)

   PROCEDURE do_release ;
    (* This routine sweeps through memory and checks for nodes with
       in_use = false. *)
    VAR
     p : s_expr ;
    BEGIN
     p := normalize(initial_heap) ;
     WHILE lower(p,heap_top) DO
      CASE p^.tag OF
       cons_node : BEGIN
                    IF NOT p^.in_use
                     THEN free_memory(p,node_size) ;
                    p := normalize(ptr(seg(p^),ofs(p^) + node_allocation)) ;
                   END ;
       free_node : BEGIN
                    block_allocation := (p^.block_cnt + 1) SHL 3 ;
                    free_memory(p,block_allocation) ;
                    p := normalize(ptr(seg(p^),ofs(p^) + block_allocation)) ;
                   END ;
       number    : BEGIN
                    IF NOT p^.in_use
                     THEN free_memory(p,number_allocation) ;
                    p := normalize(ptr(seg(p^),ofs(p^) + number_allocation)) ;
                   END ;
       symbol    : BEGIN
                    string_allocation := allocation_size(string_base +
                                                length(p^.string_data) + 1) ;
                    IF NOT p^.in_use
                     THEN free_memory(p,string_base + length(p^.string_data)
                                      + 1) ;
                    p := normalize(ptr(seg(p^),ofs(p^) + string_allocation)) ;
                   END ;
      END ;
    END ; (* do_release *)

   BEGIN
    free := NIL ;
    total_free := 0.0 ;
    heap_top := HeapPtr ;
    string_base := sizeof(node_type) + sizeof(boolean) ;
    node_allocation := allocation_size(node_size) ;
    number_allocation := allocation_size(sizeof(node_type) + sizeof(boolean) +
                                         sizeof(real)) ;
    do_release ;
   END ; (* release_mem *)

  BEGIN
   write('*') ;
   unmark_mem ;
   mark_mem(saved_list) ;
   mark_mem(ques_list) ;
   release_mem ;
   write(back_space) ;
   clreol ;
  END ; (* collect_garbage *)


 PROCEDURE test_memory ;
  (* This routine activates the garbage collector, if the the total available
     memory (free_list + heap) is less than a specified amount. Lowering the
     minimum causes garbage collection to be called less often, but if you
     make it too small you may not have enough room left for recursion or any
     temporary lists you need. Using 10000 is probably being overly
     cautious.   *)
  BEGIN
   IF (memavail * 16.0) + total_free < 10000
    THEN collect_garbage ;
  END ; (* test_memory *)


 FUNCTION yes_no : boolean ;
  (* reads yes/no answers from the key board, accepts Y,y,N,n and complains
     about anything else. It returns true, if the answer was yes. *)
  VAR
   ch : char ;
  BEGIN
   read(kbd,ch) ;
   IF ch IN ['Y','y','N','n']
    THEN
     BEGIN
      write(ch) ;
      yes_no := (ch IN ['Y','y']) ;
     END
    ELSE
     BEGIN
      write(bell) ;
      yes_no := yes_no ;
     END ;
  END ; (* yes_no *)


 FUNCTION read_string(msg : string80) : s_expr ;
  (* writes a prompt, reads a string from the keyboard and return an
     s_expr which contains the string. Each string is unique. We could
     save heap space by searching the heap for a matching string, but that
     takes too long. *)
  VAR
   s : string80 ;
  BEGIN
   writeln ;
   writeln(msg) ;
   write('=> ') ;
   readln(s) ;
   strip_leading_blanks(s) ;
   strip_trailing_blanks(s) ;
   read_string := alloc_str(s) ;
  END ; (* read_string *)


 PROCEDURE print_list(VAR f : text_file ; list : s_expr) ;
  (* recursively traverses the list and prints its elements. This is
     not a pretty printer, so the lists may look a bit messy.
     This routine is inclued for debugging purposes *)

  PROCEDURE print_tail(p : s_expr) ;
   BEGIN
    WHILE p <> NIL DO
     BEGIN
      print_list(f,car(p)) ;
      p := cdr(p) ;
      IF p <> NIL
       THEN write(f,',') ;
     END ;
   END ; (* print_tail *)

  BEGIN
   IF list <> NIL
    THEN
     CASE tag_value(list) OF
      symbol    : IF (pos(' ',string_val(list)) <> 0) OR
                     (length(string_val(list)) = 0)
                   THEN write(f,concat('''',string_val(list),
                                 ''''))
                   ELSE write(f,string_val(list)) ;
      number    : write(f,string_val(list)) ;
      cons_node : BEGIN
                   write(f,' [') ;
                   print_tail(list) ;
                   write(f,'] ') ;
                  END ;
     END ;
  END ; (* print_list *)


(* ------------------------------------------------------------------------
        End of utility routines
   ------------------------------------------------------------------------ *)


 PROCEDURE initialize ;
  BEGIN
   rule_count := 0 ;
   free := NIL ;
   rule_list := NIL ;
   ques_list := NIL ;
   saved_list := NIL ;
   clrscr ;
   writeln('Rules - Rule Builder Ver 1.0') ;
   writeln ;
   writeln('Copyright [c]    1987          Knowledge Garden Inc.') ;
   writeln('                               473A Malden Bridge Rd.') ;
   writeln('                               Nassau, NY 12123') ;
   writeln ;
   writeln ;
   mark(initial_heap)
  END ; (* initialize *)


 PROCEDURE get_rule_file ;
  (* get the file which will contain the rule base and open it *)
  VAR
   rule_name : string80 ;

  PROCEDURE open_the_file ;
   BEGIN
    assign(rule_file,rule_name) ;
    {$I-}
    rewrite(rule_file) ;
    {$I+}
    IF ioresult <> 0
     THEN
      BEGIN
       {$I-}
       close(rule_file) ;
       {$I+}
       get_rule_file ;
      END ;
   END ; (* open_the_file *)

  BEGIN
   writeln ;
   writeln('To what file should the rules be written ?') ;
   write('=> ') ;
   readln(rule_name) ;
   IF open(rule_file,rule_name)
    THEN
     BEGIN
      close(rule_file) ;
      writeln(rule_name,' already exists.') ;
      write('Do you want to overwrite it (Y/N) ? ') ;
      IF yes_no
       THEN open_the_file
       ELSE get_rule_file ;
     END
    ELSE open_the_file ;
  END ; (* get_rule_file *)


 FUNCTION get_main_goal : s_expr ;
  (* reads the main goal for the consultation. For systems like M.1 and
     Insight 2+, you will have to write the goal to the rule file. MicroExpert
     doesn't need it. main_goal is passed to build_rule. *)
  BEGIN
   clrscr ;
   writeln('RULES builds rule based classification systems.') ;
   get_main_goal := read_string('For what attribute do you want the system to determine a value ?') ;
  END ; (* get_main_goal *)


 PROCEDURE write_rules(goal,r_list : s_expr) ;
  (* writes the rules to the file, goal is the goal attribute for the
     set of rules conained in r_list. Each element of r_list is a rule,
     build_rule describes the structure of r_list. To write rules for other
     systems, change this routine. Foe M.1 and Insight 2+, the changes are
     minor. *)
  VAR
   c_list : s_expr ;
  BEGIN
   WHILE r_list <> NIL DO
    BEGIN
     rule_count := succ(rule_count) ;
     writeln(rule_file,rule_count) ;
     write(rule_file,'If   ') ;
     c_list := car(car(r_list)) ;
     WHILE c_list <> NIL DO
      BEGIN
       writeln(rule_file,string_val(car(car(c_list))),' is ',string_val(car(cdr(car(c_list))))) ;
       c_list := cdr(c_list) ;
       IF c_list <> NIL
        THEN write(rule_file,'and  ')
        ELSE write(rule_file,'then ') ;
      END ;
     writeln(rule_file,string_val(goal),' is ',string_val(car(cdr(car(r_list))))) ;
     writeln(rule_file,'.') ;
     writeln(rule_file) ;
     r_list := cdr(r_list) ;
    END ;
  END ; (* write_rules *)


 FUNCTION build_rule(goal,cond_list : s_expr) : s_expr ;
  (* This is the main routine which carries on the dialog to produce rules.
     goal is the attribute we are trying to build a decision tree for.
     cond_list is a list containing attribute value pairs which have been
     determined so far. If a value of goal can be determined from the pairs
     on cond_list, we just return cond_list and the value of goal. If
     goal cannot be determined, we need another attribute. The new attribute
     can be determined by

       1) asking a question
       2) building a set of rules to determine its value.

    For case 1, we get the question, its possible answers and call build_rule
    again for to try and build rules for each of the possible answers.
    For case 2, we call build_rule to build a set of rules which determine
    the new attribute. We then call build_rule with each possible value
    for the the new attribute from its rule set.

    When we are all done, build_rule will produce a structure similar
    to the following:

    build_rule(X,cond_list) returns

      ( (cond1 X1) (cond2 X2) (cond3 X3) ....)

      cond_list,cond1 etc. consist of attribute-value pairs,
      i.e cond1 = ( (attr1 value1) (attr2 value2) .... )

      X1, X2, X3 etc. are the values of goal. The list (cond1 X1) represents
      the rule: IF  attr1 is value1
                and attr2 is value2
                and ....
                then goal is X1.  *)
  TYPE
   binary_choice = 1 .. 2 ;
  VAR
   attrib,val_list : s_expr ;

  FUNCTION value_determined : boolean ;
   (* prints the current condition list and asks whether these conditions
      are sufficient to determine the value of goal *)

   PROCEDURE print_cond_list ;
    VAR
     p : s_expr ;
    BEGIN
     writeln('The conditions determined so far are:') ;
     writeln ;
     p := cond_list ;
     WHILE p <> NIL DO
      BEGIN
       writeln(string_val(car(car(p))),' = ',string_val(car(cdr(car(p))))) ;
       p := cdr(p) ;
      END ;
     writeln ;
    END ; (* print_cond_list *)

   BEGIN
    clrscr ;
    IF cond_list <> NIL
     THEN
      BEGIN
       print_cond_list ;
       write('Can ',string_val(goal),' be determined from these conditions (Y/N) ? ') ;
       value_determined := yes_no ;
       writeln ;
      END
     ELSE value_determined := false ;
   END ; (* value_determined *)

  FUNCTION get_value : s_expr ;
   (* read a value for goal *)
   BEGIN
    get_value := read_string(concat('Given these conditions, what is the value of ',
                                    string_val(goal),' ?')) ;
   END ; (* get_value *)

  FUNCTION get_choice : binary_choice ;
   (* reads whether this is a case 1 or 2 attribute *)

   FUNCTION read_choice : binary_choice ;
    VAR
     num : char ;
    BEGIN
     read(kbd,num) ;
     IF num IN ['1','2']
      THEN
       BEGIN
        writeln(num) ;
        read_choice := tointeger(num) ;
       END
      ELSE
       BEGIN
        write(bell) ;
        read_choice := read_choice ;
       END ;
    END ; (* read_choice *)

   BEGIN
    writeln ;
    writeln('Can ',string_val(attrib),' be determined by') ;
    writeln ;
    writeln('1) a single question') ;
    writeln('2) a series of questions') ;
    writeln ;
    write('Enter 1 or 2 => ') ;
    get_choice := read_choice ;
   END ; (* get_choice *)

  FUNCTION stored_value_list(atr : s_expr) : s_expr ;
   (* Checks ques_list to see if we already have a question and list
      of values for this attrib, the new attribute. If we do, it returns
      the value list for the attribute *)
   VAR
    atr_str : string80 ;
    p : s_expr ;
    found : boolean ;
   BEGIN
    atr_str := toupper(string_val(atr)) ;
    p := ques_list ;
    found := false ;
    WHILE (p <> NIL) AND (NOT found) DO
     IF atr_str = toupper(string_val(car(car(p))))
      THEN found := true
      ELSE p := cdr(p) ;
    IF found
     THEN stored_value_list := car(cdr(car(p)))
     ELSE stored_value_list := NIL ;
   END ; (* stored_value_list *)

  FUNCTION value_list(v_list : s_expr) : s_expr ;
   (* get the list of values and possibly the question for attrib. It
      the calls build_rule to continue building the rules for goal *)
   VAR
    q_list,r1,new_rules,new_cond_list : s_expr ;

   FUNCTION question : s_expr ;
    VAR
     s : string80 ;
     p : s_expr ;
    BEGIN
     clrscr ;
     writeln('Enter the question to be asked for ',string_val(attrib),'.') ;
     writeln('Enter a ''.'' followed by <Enter> to terminate.') ;
     writeln ;
     p := NIL ;
     readln(s) ;
     WHILE s <> '.' DO
      BEGIN
       p := append_list(p,cons(alloc_str(s),NIL)) ;
       readln(s) ;
      END ;
     writeln ;
     question := p ;
    END ; (* question *)

   FUNCTION read_values : s_expr ;
    VAR
     s : string80 ;
     p : s_expr ;
    BEGIN
     writeln ;
     writeln('Enter the possible values for ',string_val(attrib),' one per line.') ;
     writeln('Enter a ''.'' followed by <Enter> to terminate.') ;
     writeln ;
     p := NIL ;
     readln(s) ;
     WHILE s <> '.' DO
      BEGIN
       strip_leading_blanks(s) ;
       strip_trailing_blanks(s) ;
       IF s <> ''
        THEN p := cons(alloc_str(s),p) ;
       readln(s) ;
      END ;
     writeln ;
     read_values := p ;
    END ; (* read_values *)

   BEGIN
    IF v_list = NIL
     THEN
      BEGIN
       q_list := question ;
       v_list := read_values ;
       ques_list := cons(cons(attrib,cons(v_list,q_list)),ques_list) ;
      END ;
    saved_list := cons(v_list,saved_list) ;
    new_rules := NIL ;
    WHILE v_list <> NIL DO
     BEGIN
      new_cond_list := append_list(cond_list,cons(cons(attrib,cons(car(v_list),NIL)),NIL)) ;
      saved_list := cons(new_rules,saved_list) ;
      r1 := build_rule(goal,new_cond_list) ;
      saved_list := cdr(saved_list) ;
      new_rules := append_list(new_rules,r1) ;
      v_list := cdr(v_list) ;
     END ;
    saved_list := cdr(saved_list) ;
    value_list := new_rules ;
   END ; (* value_list *)

  FUNCTION build_new_tree : s_expr ;
   (* Case 2 - build a new set of rules and the use the values returned
      to finish building rules for goal *)
   VAR
    new_rules,p,r1,new_cond_list,org_rules,v_list : s_expr ;

   FUNCTION rule_conclusions : s_expr ;
    (* Extract the conclusions from the rule list *)
    VAR
     q,va_list : s_expr ;
     conc_val : string80 ;

    FUNCTION on_list(list : s_expr) : boolean ;
     BEGIN
      IF list = NIL
       THEN on_list := false
      ELSE IF conc_val = toupper(string_val(car(list)))
       THEN on_list := true
      ELSE on_list := on_list(cdr(list)) ;
     END ; (* on_list *)

    BEGIN
     q := new_rules ;
     va_list := NIL ;
     WHILE q <> NIL DO
      BEGIN
       conc_val := toupper(string_val(last(car(q)))) ;
       IF NOT on_list(va_list)
        THEN va_list := cons(last(car(q)),va_list) ;
       q := cdr(q) ;
      END ;
     rule_conclusions := va_list ;
    END ; (* rule_conclusions *)

   BEGIN
    new_rules := build_rule(attrib,NIL) ;
    saved_list := cons(new_rules,saved_list) ;
    write_rules(attrib,new_rules) ;
    v_list := rule_conclusions ;
    saved_list := cons(v_list,saved_list) ;
    org_rules := NIL ;
    p := v_list ;
    WHILE p <> NIL DO
     BEGIN
      new_cond_list := append_list(cond_list,cons(cons(attrib,cons(car(p),NIL)),NIL)) ;
      saved_list := cons(org_rules,saved_list) ;
      r1 := build_rule(goal,new_cond_list) ;
      saved_list := cdr(saved_list) ;
      org_rules := append_list(org_rules,r1) ;
      p := cdr(p) ;
     END ;
    saved_list := cdr(cdr(saved_list)) ;
    build_new_tree := org_rules ;
   END ; (* build_new_tree *)

  BEGIN
   saved_list := cons(cond_list,saved_list) ;
   test_memory ;
   IF value_determined
    THEN build_rule := cons(cons(cond_list,cons(get_value,NIL)),NIL)
    ELSE
     BEGIN
      IF cond_list = NIL
       THEN attrib := read_string(concat('In order to determine ',
                                  string_val(goal),
                                  ', what is the most important attribute ?'))
       ELSE attrib := read_string(concat('In order to find ',
                                  string_val(goal),
                                  ', what attribute should be determined next ?')) ;
      saved_list := cons(attrib,saved_list) ;
      val_list := stored_value_list(attrib) ;
      IF val_list = NIL
       THEN
        IF get_choice = 1
         THEN build_rule := value_list(NIL)
         ELSE build_rule := build_new_tree
       ELSE build_rule := value_list(val_list) ;
      saved_list := cdr(saved_list) ;
     END ;
   saved_list := cdr(saved_list) ;
  END ; (* build_rule *)


 PROCEDURE write_questions(q_list : s_expr) ;
  (* print the questions, the question list has the following format:
     ( (attrib values line1 line2 line3 ...)
       (attrib values line1 line2 ....) .....)
     values is a list of values - (val1 val2 ....),
     lines are the individual lines from the question. *)
  VAR
   p : s_expr ;
  BEGIN
   WHILE q_list <> NIL DO
    BEGIN
     writeln(rule_file,'prompt ',string_val(car(car(q_list)))) ;
     p := cdr(cdr(car(q_list))) ;
     WHILE p <> NIL DO
      BEGIN
       writeln(rule_file,string_val(car(p))) ;
       p := cdr(p) ;
      END ;
     writeln(rule_file,'.') ;
     writeln(rule_file) ;
     q_list := cdr(q_list) ;
    END ;
  END ; (* write_questions *)


 BEGIN
  initialize ;
  get_rule_file ;
  main_goal := get_main_goal ;
  saved_list := cons(main_goal,NIL) ;
  rule_list := build_rule(main_goal,NIL) ;
  write_rules(main_goal,rule_list) ;
  write_questions(ques_list) ;
  close(rule_file) ;
 END.
